"""Math utility functions for vector operations."""

import logging

import numpy as np

logger = logging.getLogger(__name__)

Matrix = list[list[float]] | list[np.ndarray] | np.ndarray


def cosine_similarity(X: Matrix, Y: Matrix) -> np.ndarray:
    """
    Compute row-wise cosine similarity between two equal-width matrices.

    Parameters
    ----------
    X : Matrix
        A matrix of shape (m, n), where `m` is the number of rows and `n` is the
        number of columns (features).
    Y : Matrix
        A matrix of shape (p, n), where `p` is the number of rows and `n` is the
        number of columns (features).

    Returns
    -------
    :
        A matrix of shape (m, p) containing the cosine similarity scores
        between each row of `X` and each row of `Y`.

    Raises
    ------
    ValueError
        If the number of columns in `X` and `Y` are not equal.

    Notes
    -----
    - If the `simsimd` library is available, it will be used for performance
      optimization. Otherwise, the function falls back to a NumPy implementation.
    - Divide-by-zero and invalid values in similarity calculations are replaced
      with 0.0 in the output.
    """
    if len(X) == 0 or len(Y) == 0:
        return np.array([])

    X = np.array(X)
    Y = np.array(Y)
    if X.shape[1] != Y.shape[1]:
        raise ValueError(
            f"Number of columns in X and Y must be the same. X has shape {X.shape} "
            f"and Y has shape {Y.shape}."
        )
    try:
        import simsimd as simd

        X = np.array(X, dtype=np.float32)
        Y = np.array(Y, dtype=np.float32)
        Z = 1 - np.array(simd.cdist(X, Y, metric="cosine"))
        return Z
    except ImportError:
        logger.debug(
            "Unable to import simsimd, defaulting to NumPy implementation. If you want "
            "to use simsimd please install with `pip install simsimd`."
        )
        X_norm = np.linalg.norm(X, axis=1)
        Y_norm = np.linalg.norm(Y, axis=1)
        # Ignore divide by zero errors run time warnings as those are handled below.
        with np.errstate(divide="ignore", invalid="ignore"):
            similarity = np.dot(X, Y.T) / np.outer(X_norm, Y_norm)
        similarity[np.isnan(similarity) | np.isinf(similarity)] = 0.0
        return similarity


def cosine_similarity_top_k(
    X: Matrix,
    Y: Matrix,
    top_k: int | None,
    score_threshold: float | None = None,
) -> tuple[list[tuple[int, int]], list[float]]:
    """
    Row-wise cosine similarity with optional top-k and score threshold filtering.

    Parameters
    ----------
    X : Matrix
        A matrix of shape (m, n), where `m` is the number of rows and `n` is the
        number of columns (features).
    Y : Matrix
        A matrix of shape (p, n), where `p` is the number of rows and `n` is the
        number of columns (features).
    top_k : int, optional
        Max number of results to return.
    score_threshold:
        Minimum score to return.

    Returns
    -------
    list[tuple[int, int]]
        Two-tuples of indices `(X_idx, Y_idx)` indicating the respective rows in
        `X` and `Y`.
    list[float]
        The corresponding cosine similarities.
    """
    if len(X) == 0 or len(Y) == 0:
        return [], []
    score_array = cosine_similarity(X, Y)
    score_threshold = score_threshold or -1.0
    score_array[score_array < score_threshold] = 0
    top_k = min(top_k or len(score_array), np.count_nonzero(score_array))
    top_k_idxs = np.argpartition(score_array, -top_k, axis=None)[-top_k:]
    top_k_idxs = top_k_idxs[np.argsort(score_array.ravel()[top_k_idxs])][::-1]
    ret_idxs = np.unravel_index(top_k_idxs, score_array.shape)
    scores = score_array.ravel()[top_k_idxs].tolist()
    return list(zip(*ret_idxs)), scores  # type: ignore



import json
import os

import pytest
from graph_retriever import Content
from graph_retriever.adapters import Adapter
from graph_retriever.adapters.in_memory import InMemory
from graph_retriever.testing import embeddings


def load_animal_content(embedding: embeddings.AnimalEmbeddings) -> list[Content]:
    documents = []

    path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "../../../../data/animals.jsonl")
    )
    with open(path) as file:
        for line in file:
            data = json.loads(line.strip())
            documents.append(
                Content(
                    id=data["id"],
                    content=data["text"],
                    embedding=embedding(data["text"]),
                    metadata=data["metadata"],
                )
            )
    return documents


@pytest.fixture(scope="session")
def animals() -> Adapter:
    embedding = embeddings.AnimalEmbeddings()
    return InMemory(embedding, load_animal_content(embedding))


ANIMALS_QUERY: str = "small agile mammal"
ANIMALS_DEPTH_0_EXPECTED: list[str] = ["fox", "mongoose"]



from __future__ import annotations

from collections.abc import Callable
from typing import Any

from pydantic import BaseModel


class Content(BaseModel):
    """
    Model representing retrieved content.

    Parameters
    ----------
    id : str
        The ID of the content.
    content : str
        The content.
    embedding : list[float]
        The embedding of the content.
    score : float | None
        The similarity of the embedding to the query.
        This is optional, and may not be set depending on the content.
    metadata : dict[str, Any]
        The metadata associated with the content.
    mime_type : str
        The MIME type of the content.
    """

    id: str
    content: str
    embedding: list[float]
    metadata: dict[str, Any] = {}

    mime_type: str = "text/plain"
    score: float | None = None

    @staticmethod
    def new(
        id: str,
        content: str,
        embedding: list[float] | Callable[[str], list[float]],
        score: float | None = None,
        metadata: dict[str, Any] = {},
        mime_type: str = "text/plain",
    ) -> Content:
        """
        Create a new content.

        Parameters
        ----------
        id : str
            The ID of the content.
        content : str
            The content.
        embedding : list[float] | Callable[[str], list[float]]
            The embedding, or a function to apply to the content to compute the
            embedding.
        score : float | None, optional
            The similarity of the embedding to the query.
        metadata : dict[str, Any], optional
            The metadata associated with the content.
        mime_type : str, optional
            The MIME type of the content.

        Returns
        -------
        :
            The created content.
        """
        return Content(
            id=id,
            content=content,
            embedding=embedding(content) if callable(embedding) else embedding,
            score=score,
            metadata=metadata,
            mime_type=mime_type,
        )



"""Implements the traversal logic for graph-based document retrieval."""

from collections.abc import Iterable, Sequence
from typing import Any

from graph_retriever.adapters import Adapter
from graph_retriever.content import Content
from graph_retriever.edges.metadata import EdgeSpec, MetadataEdgeFunction
from graph_retriever.strategies import Strategy
from graph_retriever.types import Edge, EdgeFunction, Node


def traverse(
    query: str,
    *,
    edges: list[EdgeSpec] | EdgeFunction,
    strategy: Strategy,
    store: Adapter,
    metadata_filter: dict[str, Any] | None = None,
    initial_root_ids: Sequence[str] = (),
    store_kwargs: dict[str, Any] = {},
) -> list[Node]:
    """
    Perform a graph traversal to retrieve nodes for a specific query.

    Parameters
    ----------
    query : str
        The query string for the traversal.
    edges : list[EdgeSpec] | EdgeFunction
        A list of [](`~graph_retriever.EdgeSpec`) for use in creating a
        [](`~graph_retriever.edges.metadata.MetadataEdgeFunction`),
        or an [](`~graph_retriever.EdgeFunction`).
    strategy : Strategy
        The traversal strategy that defines how nodes are discovered, selected,
        and finalized.
    store : Adapter
        The vector store adapter used for similarity searches and document
        retrieval.
    metadata_filter : dict[str, Any], optional
        Optional filter for metadata during traversal.
    initial_root_ids : Sequence[str], optional
        IDs of the initial root nodes for the traversal.
    store_kwargs : dict[str, Any], optional
        Additional arguments passed to the store adapter.

    Returns
    -------
    :
        Nodes returned by the traversal.
    """
    traversal = _Traversal(
        query=query,
        edges=edges,
        strategy=strategy.model_copy(deep=True),
        store=store,
        metadata_filter=metadata_filter,
        initial_root_ids=initial_root_ids,
        store_kwargs=store_kwargs,
    )
    return traversal.traverse()


async def atraverse(
    query: str,
    *,
    edges: list[EdgeSpec] | EdgeFunction,
    strategy: Strategy,
    store: Adapter,
    metadata_filter: dict[str, Any] | None = None,
    initial_root_ids: Sequence[str] = (),
    store_kwargs: dict[str, Any] = {},
) -> list[Node]:
    """
    Asynchronously perform a graph traversal to retrieve nodes for a specific query.

    Parameters
    ----------
    query : str
        The query string for the traversal.
    edges : list[EdgeSpec] | EdgeFunction
        A list of [](`~graph_retriever.EdgeSpec`) for use in creating a
        [](`~graph_retriever.edges.metadata.MetadataEdgeFunction`),
        or an [](`~graph_retriever.EdgeFunction`).
    strategy : Strategy
        The traversal strategy that defines how nodes are discovered, selected,
        and finalized.
    store : Adapter
        The vector store adapter used for similarity searches and document
        retrieval.
    metadata_filter : dict[str, Any], optional
        Optional filter for metadata during traversal.
    initial_root_ids : Sequence[str], optional
        IDs of the initial root nodes for the traversal.
    store_kwargs : dict[str, Any], optional
        Additional arguments passed to the store adapter.

    Returns
    -------
    :
        Nodes returned by the traversal.
    """
    traversal = _Traversal(
        query=query,
        edges=edges,
        strategy=strategy.model_copy(deep=True),
        store=store,
        metadata_filter=metadata_filter,
        initial_root_ids=initial_root_ids,
        store_kwargs=store_kwargs,
    )
    return await traversal.atraverse()


class _Traversal:
    """
    Handles a single traversal operation for a graph-based retrieval system.

    The `_Traversal` class manages the process of discovering, visiting, and selecting
    nodes within a graph, based on a query and a traversal strategy. It supports
    synchronous and asynchronous traversal, enabling retrieval of documents in a
    controlled, iterative manner.

    This class should not be reused between traversals.
    """

    def __init__(
        self,
        query: str,
        *,
        edges: list[EdgeSpec] | EdgeFunction,
        strategy: Strategy,
        store: Adapter,
        metadata_filter: dict[str, Any] | None = None,
        initial_root_ids: Sequence[str] = (),
        store_kwargs: dict[str, Any] = {},
    ) -> None:
        self.query = query

        self.edge_function: EdgeFunction
        if isinstance(edges, list):
            self.edge_function = MetadataEdgeFunction(edges)
        elif callable(edges):
            self.edge_function = edges
        else:
            raise ValueError(f"Invalid edges: {edges}")

        self.strategy = strategy
        self.store = store
        self.metadata_filter = metadata_filter
        self.initial_root_ids = initial_root_ids
        self.store_kwargs = store_kwargs

        self._used = False
        self._visited_edges: set[Edge] = set()
        self._edge_depths: dict[Edge, int] = {}
        self._node_cache: dict[str, Node] = {}
        self._selected_nodes: dict[str, Node] = {}

    def _check_first_use(self):
        assert not self._used, "Traversals cannot be re-used."
        self._used = True

    def traverse(self) -> list[Node]:
        """
        Execute the traversal synchronously.

        This method retrieves initial candidates, discovers and visits nodes,
        and explores edges iteratively until the traversal is complete.

        Returns
        -------
        :
            The final set of nodes resulting from the traversal.
        """
        self._check_first_use()

        # Retrieve initial candidates.
        initial_docs = self._fetch_initial_candidates()
        self.add_docs(initial_docs, depth=0)

        if self.initial_root_ids:
            neighborhood_adjacent_docs = self._fetch_neighborhood_candidates()
            self.add_docs(neighborhood_adjacent_docs, depth=0)

        while True:
            # Select the next batch of nodes, and (new) outgoing edges.
            next_outgoing_edges = self.select_next_edges()
            if next_outgoing_edges is None:
                break
            elif next_outgoing_edges:
                # Find the (new) document with incoming edges from those edges.
                adjacent_docs = self._fetch_adjacent(next_outgoing_edges)
                self.add_docs(adjacent_docs)

        return self.finish()

    async def atraverse(self) -> list[Node]:
        """
        Execute the traversal asynchronously.

        This method retrieves initial candidates, discovers and visits nodes,
        and explores edges iteratively until the traversal is complete.

        Returns
        -------
        :
            The final set of nodes resulting from the traversal.
        """
        self._check_first_use()

        # Retrieve initial candidates.
        initial_docs = await self._afetch_initial_candidates()
        self.add_docs(initial_docs, depth=0)

        if self.initial_root_ids:
            neighborhood_adjacent_docs = await self._afetch_neighborhood_candidates()
            self.add_docs(neighborhood_adjacent_docs, depth=0)

        while True:
            # Select the next batch of nodes, and (new) outgoing edges.
            next_outgoing_edges = self.select_next_edges()
            if next_outgoing_edges is None:
                break
            elif next_outgoing_edges:
                # Find the (new) document with incoming edges from those edges.
                adjacent_docs = await self._afetch_adjacent(next_outgoing_edges)
                self.add_docs(adjacent_docs)

        return self.finish()

    def _fetch_initial_candidates(self) -> list[Content]:
        """
        Retrieve initial candidates based on the query and strategy.

        Returns
        -------
        :
            The initial content retrieved via similarity search.
        """
        query_embedding, docs = self.store.search_with_embedding(
            query=self.query,
            k=self.strategy.start_k,
            filter=self.metadata_filter,
            **self.store_kwargs,
        )
        self.strategy._query_embedding = query_embedding
        return docs

    async def _afetch_initial_candidates(self) -> list[Content]:
        query_embedding, docs = await self.store.asearch_with_embedding(
            query=self.query,
            k=self.strategy.start_k,
            filter=self.metadata_filter,
            **self.store_kwargs,
        )
        self.strategy._query_embedding = query_embedding
        return docs

    def _fetch_neighborhood_candidates(self) -> Iterable[Content]:
        """
        Retrieve neighborhood candidates for traversal.

        This method fetches initial root documents, converts them to nodes, and records
        their outgoing edges as visited. It then fetches additional candidates adjacent
        to these nodes.

        Returns
        -------
        :
            The set of contents adjacent to the initial root nodes.
        """
        neighborhood_docs = self.store.get(self.initial_root_ids)
        neighborhood_nodes = self.add_docs(neighborhood_docs)

        # Record the neighborhood nodes (specifically the outgoing edges from the
        # neighborhood) as visited.
        outgoing_edges = self.visit_nodes(neighborhood_nodes.values())

        # Fetch the candidates.
        return self._fetch_adjacent(outgoing_edges)

    async def _afetch_neighborhood_candidates(
        self,
    ) -> Iterable[Content]:
        """
        Asynchronously retrieve neighborhood candidates for traversal.

        This method fetches initial root documents, converts them to nodes, and records
        their outgoing edges as visited. It then fetches additional candidates adjacent
        to these nodes.

        Returns
        -------
        :
            The set of contents adjacent to the initial root nodes.
        """
        neighborhood_docs = await self.store.aget(self.initial_root_ids)
        neighborhood_nodes = self.add_docs(neighborhood_docs)

        # Record the neighborhood nodes (specifically the outgoing edges from the
        # neighborhood) as visited.
        outgoing_edges = self.visit_nodes(neighborhood_nodes.values())

        # Fetch the candidates.
        return await self._afetch_adjacent(outgoing_edges)

    def _fetch_adjacent(self, edges: set[Edge]) -> Iterable[Content]:
        """
        Retrieve documents adjacent to the specified outgoing edges.

        This method uses the vector store adapter to fetch documents connected to
        the provided edges.

        Parameters
        ----------
        edges : set[Edge]
            The edges whose adjacent documents need to be fetched.

        Returns
        -------
        :
            The set of content adjacent to the specified edges.
        """
        return self.store.adjacent(
            edges=edges,
            query_embedding=self.strategy._query_embedding,
            k=self.strategy.adjacent_k,
            filter=self.metadata_filter,
            **self.store_kwargs,
        )

    async def _afetch_adjacent(self, edges: set[Edge]) -> Iterable[Content]:
        """
        Asynchronously retrieve documents adjacent to the specified outgoing edges.

        This method uses the vector store adapter to fetch documents connected to
        the provided edges.

        Parameters
        ----------
        edges : set[Edge]
            The edges whose adjacent documents need to be fetched.

        Returns
        -------
        :
            The set of content adjacent to the specified edges.
        """
        return await self.store.aadjacent(
            edges=edges,
            query_embedding=self.strategy._query_embedding,
            k=self.strategy.adjacent_k,
            filter=self.metadata_filter,
            **self.store_kwargs,
        )

    def _content_to_new_node(
        self, content: Content, *, depth: int | None = None
    ) -> Node | None:
        """
        Convert a content into a new node for the traversal.

        This method checks whether the document has already been processed. If not,
        it creates a new `Node` instance, associates it with the document's metadata,
        and calculates its depth based on the incoming edges.

        Parameters
        ----------
        content : Content
            The content to convert into a node.
        depth : int, optional
            The depth of the node. If None, the depth is calculated based on the
            incoming edges.

        Returns
        -------
        :
            The newly created node, or None if the document has already been
            processed.
        """
        if content.id in self._node_cache:
            return None

        # Determine incoming/outgoing edges.
        edges = self.edge_function(content)

        # Compute the depth
        if depth is None:
            depth = min(
                [
                    d
                    for e in edges.incoming
                    if (d := self._edge_depths.get(e, None)) is not None
                ],
                default=0,
            )

        node = Node(
            id=content.id,
            content=content.content,
            depth=depth,
            embedding=content.embedding,
            metadata=content.metadata,
            incoming_edges=edges.incoming,
            outgoing_edges=edges.outgoing,
        )

        self._node_cache[node.id] = node

        return node

    def add_docs(
        self, contents: Iterable[Content], *, depth: int | None = None
    ) -> dict[str, Node]:
        """
        Add a batch of content to the traversal and convert them into nodes.

        This method records the depth of new nodes, filters them based on the
        strategy's maximum depth, and updates the strategy with the discovered nodes.

        Parameters
        ----------
        contents : Iterable[Content]
            The contents to add.
        depth : int, optional
            The depth to assign to the nodes. If None, the depth is inferred
            based on the incoming edges.

        Returns
        -------
        :
            A dictionary of node IDs to the newly created nodes.
        """
        # Record the depth of new nodes.
        nodes = {
            node.id: node
            for c in contents
            if (node := self._content_to_new_node(c, depth=depth)) is not None
            if (
                self.strategy.max_depth is None or node.depth <= self.strategy.max_depth
            )
        }
        self.strategy.discover_nodes(nodes)
        return nodes

    def visit_nodes(self, nodes: Iterable[Node]) -> set[Edge]:
        """
        Mark nodes as visited and return their new outgoing edges.

        This method updates the traversal state by marking the provided nodes as visited
        and recording their outgoing edges. Outgoing edges that have not been visited
        before are identified and added to the set of edges to explore in subsequent
        traversal steps.

        Parameters
        ----------
        nodes : Iterable[Node]
            The nodes to mark as visited.

        Returns
        -------
        :
            The set of new outgoing edges that need to be explored.

        Notes
        -----
        - The `new_outgoing_edges` dictionary tracks the depth of each outgoing
        edge.
        - If a node's outgoing edge leads to a lower depth, the edge's depth is
        updated to reflect the shortest path.
        - The `_visited_edges` set is updated to include all outgoing edges
        from the provided nodes.
        """
        new_outgoing_edges: dict[Edge, int] = {}
        for node in nodes:
            node_new_outgoing_edges = node.outgoing_edges - self._visited_edges
            for edge in node_new_outgoing_edges:
                depth = new_outgoing_edges.setdefault(edge, node.depth + 1)
                if node.depth + 1 < depth:
                    new_outgoing_edges[edge] = node.depth + 1

        self._edge_depths.update(new_outgoing_edges)

        new_outgoing_edge_set = set(new_outgoing_edges.keys())
        self._visited_edges.update(new_outgoing_edge_set)
        return new_outgoing_edge_set

    def select_next_edges(self) -> set[Edge] | None:
        """
        Select the next set of edges to explore.

        This method uses the traversal strategy to select the next batch of nodes
        and identifies new outgoing edges for exploration.

        Returns
        -------
        :
            The set of new edges to explore, or None if the traversal is
            complete.
        """
        remaining = self.strategy.k - len(self._selected_nodes)

        if remaining <= 0:
            return None

        next_nodes = self.strategy.select_nodes(limit=remaining)
        if not next_nodes:
            return None

        next_nodes = [n for n in next_nodes if n.id not in self._selected_nodes]
        if len(next_nodes) == 0:
            return None

        self._selected_nodes.update({n.id: n for n in next_nodes})
        new_outgoing_edges = self.visit_nodes(next_nodes)
        return new_outgoing_edges

    def finish(self) -> list[Node]:
        """
        Finalize the traversal and return the final set of nodes.

        This method finalizes the selected nodes using the traversal strategy,
        processes their metadata, and assembles the final list of nodes.

        Returns
        -------
        :
            The final set of nodes resulting from the traversal.
        """
        return list(self.strategy.finalize_nodes(self._selected_nodes.values()))



import pytest
from graph_retriever.strategies import (
    Eager,
    Mmr,
    Strategy,
)


def test_build_strategy_base():
    base_strategy = Eager(k=6, start_k=5, adjacent_k=9, max_depth=2)

    # base strategy with no changes
    strategy = Strategy.build(base_strategy=base_strategy)
    assert strategy == base_strategy

    # base strategy with changed k
    strategy = Strategy.build(base_strategy=base_strategy, k=7)
    assert strategy == Eager(k=7, start_k=5, adjacent_k=9, max_depth=2)

    # base strategy with invalid kwarg
    with pytest.warns(UserWarning, match=r"Unsupported key\(s\) 'invalid_kwarg' set."):
        strategy = Strategy.build(base_strategy=base_strategy, invalid_kwarg=4)
        assert strategy == base_strategy


def test_build_strategy_base_override():
    base_strategy = Eager(k=6, start_k=5, adjacent_k=9, max_depth=2)
    override_strategy = Eager(k=7, start_k=4, adjacent_k=8, max_depth=3)

    # override base strategy
    strategy = Strategy.build(
        base_strategy=base_strategy, strategy=override_strategy, k=4
    )
    assert strategy == override_strategy.model_copy(update={"k": 4})

    # override base strategy and change params
    strategy = Strategy.build(
        base_strategy=base_strategy, strategy=override_strategy, k=3, adjacent_k=7
    )
    assert strategy == Eager(k=3, start_k=4, adjacent_k=7, max_depth=3)

    # override base strategy and invalid kwarg
    with pytest.warns(UserWarning, match=r"Unsupported key\(s\) 'invalid_kwarg' set."):
        strategy = Strategy.build(
            base_strategy=base_strategy,
            strategy=override_strategy,
            k=4,
            invalid_kwarg=4,
        )
        assert strategy == override_strategy.model_copy(update={"k": 4})

    # attempt override base strategy with dict
    with pytest.raises(ValueError, match="Unsupported 'strategy'"):
        strategy = Strategy.build(
            base_strategy=base_strategy,
            strategy={"k": 9, "start_k": 7, "adjacent_k": 11},
        )


def test_build_strategy_base_override_mmr():
    base_strategy = Eager(k=6, start_k=5, adjacent_k=9, max_depth=2)
    override_strategy = Mmr(k=7, start_k=4, adjacent_k=8, max_depth=3, lambda_mult=0.3)

    # override base strategy with mmr kwarg
    with pytest.warns(UserWarning, match=r"Unsupported key\(s\) 'lambda_mult' set."):
        strategy = Strategy.build(base_strategy=base_strategy, lambda_mult=0.2)
        assert strategy == base_strategy

    # override base strategy with mmr strategy
    strategy = Strategy.build(
        base_strategy=base_strategy, strategy=override_strategy, k=4
    )
    assert strategy == override_strategy.model_copy(update={"k": 4})

    # override base strategy with mmr strategy and mmr arg
    strategy = Strategy.build(
        base_strategy=base_strategy, strategy=override_strategy, k=4, lambda_mult=0.2
    )
    assert strategy == Mmr(k=4, start_k=4, adjacent_k=8, max_depth=3, lambda_mult=0.2)

    # start with override strategy, change to base, try to set mmr arg
    with pytest.warns(UserWarning, match=r"Unsupported key\(s\) 'lambda_mult' set."):
        Strategy.build(
            base_strategy=override_strategy, strategy=base_strategy, lambda_mult=0.2
        )



import abc
import dataclasses
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

import pytest

from graph_retriever import Adapter, Edge, IdEdge, MetadataEdge
from graph_retriever.content import Content


def assert_valid_result(content: Content):
    """Assert the content is valid."""
    assert isinstance(content.id, str)
    assert_is_embedding(content.embedding)


def assert_is_embedding(value: Any):
    """Assert the value is an embedding."""
    assert isinstance(value, list)
    for item in value:
        assert isinstance(item, float)


def assert_valid_results(docs: Iterable[Content]):
    """Assert all of the contents are valid results."""
    for doc in docs:
        assert_valid_result(doc)


def assert_ids_any_order(
    results: Iterable[Content],
    expected: list[str],
) -> None:
    """Assert the results are valid and match the IDs."""
    assert_valid_results(results)

    result_ids = [r.id for r in results]
    assert set(result_ids) == set(expected), "should contain exactly expected IDs"


@dataclass
class GetCase:
    """A test case for `get` and `aget`."""

    id: str
    request: list[str]
    expected: list[str]
    filter: dict[str, Any] | None = None


GET_CASES: list[GetCase] = [
    # Currently, this is not required for `get` implementations since the
    # traversal skips making `get` calls with no IDs. Some stores (such as chroma)
    # fail in this case.
    # GetCase("none", [], []),
    GetCase("one", ["boar"], ["boar"]),
    GetCase("many", ["boar", "chinchilla", "cobra"], ["boar", "chinchilla", "cobra"]),
    GetCase(
        "missing",
        ["boar", "chinchilla", "unicorn", "cobra"],
        ["boar", "chinchilla", "cobra"],
    ),
    GetCase(
        "duplicate",
        ["boar", "chinchilla", "boar", "cobra"],
        ["boar", "chinchilla", "cobra"],
    ),
    GetCase(
        "filtered",
        ["boar", "chinchilla", "boar", "cobra"],
        ["chinchilla"],
        filter={"keywords": "andes"},
    ),
]


@dataclass
class SearchCase:
    """A test case for `similarity_search_*` and `asimilarity_search_*` methods."""

    id: str
    query: str
    expected: list[str]
    k: int | None = None
    filter: dict[str, str] | None = None

    skips: dict[str, str] = dataclasses.field(default_factory=dict)

    @property
    def kwargs(self):
        """Return keyword arguments for the test invocation."""
        kwargs = {}
        if self.k is not None:
            kwargs["k"] = self.k
        if self.filter is not None:
            kwargs["filter"] = self.filter
        return kwargs


SEARCH_CASES: list[SearchCase] = [
    SearchCase("basic", "domesticated hunters", ["cat", "horse", "chicken", "llama"]),
    SearchCase("k2", "domesticated hunters", k=2, expected=["cat", "horse"]),
    # Many stores fail in this case. Generally it doesn't happen in the code, since
    # no IDs means we don't need to make the call. Not currently part of the contract.
    # SimilaritySearchCase("k0", "domesticated hunters", k=0, expected=[]),
    SearchCase(
        "value_filter",
        "domesticated hunters",
        filter={"type": "mammal"},
        expected=["cat", "dog", "horse", "llama"],
    ),
    SearchCase(
        "list_filter",
        "domesticated hunters",
        filter={"keywords": "hunting"},
        expected=["cat"],
    ),
    SearchCase(
        "two_filters",
        "domesticated hunters",
        filter={"type": "mammal", "diet": "carnivorous"},
        expected=["cat", "dingo", "ferret"],
    ),
    # OpenSearch supports filtering on multiple values, but it is not currently
    # relied on. Since no other adapters support it, we don't test it nor should
    # traversal depend on it.
    # SimilaritySearchCase(
    #   "multi_list_filter",
    #   "domesticated hunters",
    #   filter={"keywords": ["hunting", "agile"]},
    #   expected=["cat", "fox", "gazelle", "mongoose"]
    # ),
]


@dataclass
class AdjacentCase:
    """A test case for `get_adjacent` and `aget_adjacent`."""

    id: str
    query: str
    edges: set[Edge]
    expected: list[str]

    k: int = 4
    filter: dict[str, Any] | None = None


ADJACENT_CASES: list[AdjacentCase] = [
    AdjacentCase(
        "one_edge",
        "domesticated hunters",
        edges={MetadataEdge("type", "mammal")},
        expected=["horse", "llama", "dog", "cat"],
    ),
    AdjacentCase(
        "two_edges_same_field",
        "domesticated hunters",
        edges={
            MetadataEdge("type", "mammal"),
            MetadataEdge("type", "crustacean"),
        },
        expected=[
            "cat",
            "dog",
            "horse",
            "llama",
        ],
    ),
    AdjacentCase(
        "one_ids",
        "domesticated hunters",
        edges={
            IdEdge("cat"),
        },
        expected=[
            "cat",
        ],
    ),
    AdjacentCase(
        "many_ids",
        "domesticated hunters",
        edges={
            IdEdge("cat"),
            IdEdge("dog"),
            IdEdge("unicorn"),
            IdEdge("crab"),
        },
        expected=[
            "cat",
            "dog",
            "crab",
        ],
    ),
    AdjacentCase(
        "ids_limit_k",
        "domesticated hunters",
        edges={
            IdEdge("cat"),
            IdEdge("dog"),
            IdEdge("unicorn"),
            IdEdge("antelope"),
        },
        k=2,
        expected=[
            "cat",
            "dog",
        ],
    ),
    AdjacentCase(
        "filtered_ids",
        "domesticated hunters",
        edges={
            IdEdge("boar"),
            IdEdge("chinchilla"),
            IdEdge("unicorn"),
            IdEdge("griaffe"),
        },
        filter={"keywords": "andes"},
        expected=[
            "chinchilla",
        ],
    ),
    AdjacentCase(
        "metadata_and_id",
        "domesticated hunters",
        edges={
            IdEdge("cat"),
            MetadataEdge("type", "reptile"),
        },
        k=6,
        expected=[
            "alligator",  # reptile
            "crocodile",  # reptile
            "cat",  # by ID
            "chameleon",  # reptile
            "gecko",  # reptile
            "komodo dragon",  # reptile
        ],
    ),
]


class AdapterComplianceSuite(abc.ABC):
    """
    Test suite for adapter compliance.

    To use this, create a sub-class containing a `@pytest.fixture` named
    `adapter` which returns an `Adapter` with the documents from `animals.jsonl`
    loaded.
    """

    def skip_case(self, method: str, case_id: str) -> None:
        """
        Override to skip a specific method or test case.

        Call `pytest.skip(reason)` if necessary with the reason to skip,
        or `pytest.xfail(reason)` if a failure is expected.

        Parameters
        ----------
        method : str
            The method being tested. For instance, `get`, `aget`, or
            `similarity_search_with_embedding`, etc.
        case_id : str
            The ID of the case being executed. For instance, `one` or `many`.
        """

    @pytest.fixture(params=GET_CASES, ids=lambda c: c.id)
    def get_case(self, request) -> GetCase:
        """Fixture providing the `get` and `aget` test cases."""
        return request.param

    @pytest.fixture(params=ADJACENT_CASES, ids=lambda c: c.id)
    def adjacent_case(self, request) -> AdjacentCase:
        """Fixture providing the `get_adjacent` and `aget_adjacent` test cases."""
        return request.param

    @pytest.fixture(params=SEARCH_CASES, ids=lambda c: c.id)
    def search_case(self, request) -> SearchCase:
        """Fixture providing the `(a)?similarity_search_*` test cases."""
        return request.param

    def test_get(self, adapter: Adapter, get_case: GetCase) -> None:
        """Run tests for `get`."""
        self.skip_case("get", get_case.id)
        results = adapter.get(get_case.request, filter=get_case.filter)
        assert_ids_any_order(results, get_case.expected)

    async def test_aget(self, adapter: Adapter, get_case: GetCase) -> None:
        """Run tests for `aget`."""
        self.skip_case("aget", get_case.id)
        results = await adapter.aget(get_case.request, filter=get_case.filter)
        assert_ids_any_order(results, get_case.expected)

    def test_search_with_embedding(
        self, adapter: Adapter, search_case: SearchCase
    ) -> None:
        """Run tests for `search_with_embedding`."""
        self.skip_case("search_with_embedding", search_case.id)
        embedding, results = adapter.search_with_embedding(
            search_case.query, **search_case.kwargs
        )
        assert_is_embedding(embedding)
        assert_ids_any_order(results, search_case.expected)

    async def test_asearch_with_embedding(
        self, adapter: Adapter, search_case: SearchCase
    ) -> None:
        """Run tests for `asearch_with_embedding`."""
        self.skip_case("asearch_with_embedding", search_case.id)
        embedding, results = await adapter.asearch_with_embedding(
            search_case.query, **search_case.kwargs
        )
        assert_is_embedding(embedding)
        assert_ids_any_order(results, search_case.expected)

    def test_search(self, adapter: Adapter, search_case: SearchCase) -> None:
        """Run tests for `search`."""
        self.skip_case("search", search_case.id)
        embedding = adapter.embed_query(search_case.query)
        results = adapter.search(embedding, **search_case.kwargs)
        assert_ids_any_order(results, search_case.expected)

    async def test_asearch(self, adapter: Adapter, search_case: SearchCase) -> None:
        """Run tests for `asearch`."""
        self.skip_case("asearch", search_case.id)
        embedding = adapter.embed_query(search_case.query)
        results = await adapter.asearch(embedding, **search_case.kwargs)
        assert_ids_any_order(results, search_case.expected)

    async def test_adjacent(
        self, adapter: Adapter, adjacent_case: AdjacentCase
    ) -> None:
        """Run tests for `adjacent."""
        self.skip_case("adjacent", adjacent_case.id)
        embedding = adapter.embed_query(adjacent_case.query)
        results = adapter.adjacent(
            edges=adjacent_case.edges,
            query_embedding=embedding,
            k=adjacent_case.k,
            filter=adjacent_case.filter,
        )
        assert_ids_any_order(results, adjacent_case.expected)

    async def test_aadjacent(
        self, adapter: Adapter, adjacent_case: AdjacentCase
    ) -> None:
        """Run tests for `aadjacent."""
        self.skip_case("aadjacent", adjacent_case.id)
        embedding = adapter.embed_query(adjacent_case.query)
        results = await adapter.aadjacent(
            edges=adjacent_case.edges,
            query_embedding=embedding,
            k=adjacent_case.k,
            filter=adjacent_case.filter,
        )
        assert_ids_any_order(results, adjacent_case.expected)



import heapq
from collections.abc import Callable, Iterable

from typing_extensions import override

from graph_retriever.strategies.base import Strategy
from graph_retriever.types import Node


class _ScoredNode:
    def __init__(self, score: float, node: Node) -> None:
        self.score = score
        self.node = node

    def __lt__(self, other) -> bool:
        return other.score < self.score


class Scored(Strategy):
    """Strategy selecing nodes using a scoring function."""

    scorer: Callable[[Node], float]
    per_iteration_limit: int | None = None

    _nodes: list[_ScoredNode] = []

    @override
    def discover_nodes(self, nodes: dict[str, Node]) -> None:
        for node in nodes.values():
            heapq.heappush(self._nodes, _ScoredNode(self.scorer(node), node))

    @override
    def select_nodes(self, *, limit: int) -> Iterable[Node]:
        if self.per_iteration_limit and self.per_iteration_limit < limit:
            limit = self.per_iteration_limit

        selected = []
        for _x in range(limit):
            if not self._nodes:
                break

            selected.append(heapq.heappop(self._nodes).node)
        return selected



"""Defines the base class for vector store adapters."""

import abc
import asyncio
from collections.abc import Iterable, Sequence
from typing import Any

from graph_retriever.content import Content
from graph_retriever.types import Edge, IdEdge, MetadataEdge
from graph_retriever.utils.run_in_executor import run_in_executor
from graph_retriever.utils.top_k import top_k


class Adapter(abc.ABC):
    """
    Base adapter for integrating vector stores with the graph retriever system.

    This class provides a foundation for custom adapters, enabling consistent
    interaction with various vector store implementations.
    """

    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    def embed_query(self, query: str) -> list[float]:
        """Return the embedding of the query."""
        ...

    def search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Content]]:
        """
        Return content items most similar to the query.

        Also returns the embedded query vector.

        Parameters
        ----------
        query : str
            Input text.
        k : int, default 4
            Number of content items to return.
        filter : dict[str, Any], optional
            Filter on the metadata to apply.
        kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        query_embedding : list[float]
            The query embedding used for selecting the most relevant content.
        contents : list[Content]
            List of up to `k` content items most similar to the query vector.
        """
        query_embedding = self.embed_query(query)
        docs = self.search(
            embedding=query_embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        return query_embedding, docs

    async def asearch_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Content]]:
        """
        Asynchronously return content items most similar to the query.

        Also returns the embedded query vector.

        Parameters
        ----------
        query : str
            Input text.
        k : int, default 4
            Number of content items to return.
        filter : dict[str, Any], optional
            Filter on the metadata to apply.
        kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        query_embedding : list[float]
            The query embedding used for selecting the most relevant content.
        contents : list[Content]
            List of up to `k` content items most similar to the query
            vector.
        """
        return await run_in_executor(
            None, self.search_with_embedding, query, k, filter, **kwargs
        )

    @abc.abstractmethod
    def search(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        """
        Return content items most similar to the query vector.

        Parameters
        ----------
        embedding : list[float]
            The query embedding used for selecting the most relevant content.
        k : int, default 4
            Number of content items to return.
        filter : dict[str, Any], optional
            Filter on the metadata to apply.
        kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        :
            List of content items most similar to the query vector.
        """
        ...

    async def asearch(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        """
        Asynchronously return content items most similar to the query vector.

        Parameters
        ----------
        embedding : list[float]
            The query embedding used for selecting the most relevant content.
        k : int, default 4
            Number of content items to return.
        filter : dict[str, Any], optional
            Filter on the metadata to apply.
        kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        :
            List of content items most similar to the query vector.
        """
        return await run_in_executor(
            None,
            self.search,
            embedding,
            k,
            filter,
            **kwargs,
        )

    @abc.abstractmethod
    def get(
        self,
        ids: Sequence[str],
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        """
        Get content items by ID.

        Fewer content items may be returned than requested if some IDs are
        not found or if there are duplicated IDs. This method should **NOT**
        raise exceptions if no content items are found for some IDs.

        Users should not assume that the order of the returned content items
        matches  the order of the input IDs. Instead, users should rely on
        the ID field of the returned content items.

        Parameters
        ----------
        ids : Sequence[str]
            List of IDs to get.
        filter : dict[str, Any], optional
            Filter on the metadata to apply.
        kwargs : dict, optional
            Additional keyword arguments. These are up to the implementation.

        Returns
        -------
        :
            List of content items that were found.
        """
        ...

    async def aget(
        self,
        ids: Sequence[str],
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        """
        Asynchronously get content items by ID.

        Fewer content items may be returned than requested if some IDs are
        not found or if there are duplicated IDs. This method should **NOT**
        raise exceptions if no content items are found for some IDs.

        Users should not assume that the order of the returned content items
        matches  the order of the input IDs. Instead, users should rely on
        the ID field of the returned content items.

        Parameters
        ----------
        ids : Sequence[str]
            List of IDs to get.
        filter : dict[str, Any], optional
            Filter on the metadata to apply.
        kwargs : dict, optional
            Additional keyword arguments. These are up to the implementation.

        Returns
        -------
        :
            List of content items that were found.
        """
        return await run_in_executor(
            None,
            self.get,
            ids,
            filter,
            **kwargs,
        )

    def adjacent(
        self,
        edges: set[Edge],
        query_embedding: list[float],
        k: int,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Content]:
        """
        Return the content items with at least one matching incoming edge.

        Parameters
        ----------
        edges : set[Edge]
            The edges to look for.
        query_embedding : list[float]
            The query embedding used for selecting the most relevant content.
        k : int
            The number of relevant content items to select.
        filter : dict[str, Any], optional
            Optional metadata to filter the results.
        kwargs : dict, optional
            Keyword arguments to pass to the similarity search.

        Returns
        -------
        :
            Iterable of adjacent content items.

        Raises
        ------
        ValueError
            If unsupported edge types are encountered.
        """
        results: list[Content] = []

        ids = []
        for edge in edges:
            if isinstance(edge, MetadataEdge):
                docs = self.search(
                    embedding=query_embedding,
                    k=k,
                    filter=self._metadata_filter(base_filter=filter, edge=edge),
                    **kwargs,
                )
                results.extend(docs)
            elif isinstance(edge, IdEdge):
                ids.append(edge.id)
            else:
                raise ValueError(f"Unsupported edge: {edge}")

        if ids:
            results.extend(self.get(ids, filter=filter))

        return top_k(
            results,
            embedding=query_embedding,
            k=k,
        )

    async def aadjacent(
        self,
        edges: set[Edge],
        query_embedding: list[float],
        k: int,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Content]:
        """
        Asynchronously return the content items with at least one matching edge.

        Parameters
        ----------
        edges : set[Edge]
            The edges to look for.
        query_embedding : list[float]
            The query embedding used for selecting the most relevant content.
        k : int
            The number of relevant content items to select for the edges.
        filter : dict[str, Any], optional
            Optional metadata to filter the results.
        kwargs : dict, optional
            Keyword arguments to pass to the similarity search.

        Returns
        -------
        :
            Iterable of adjacent content items.

        Raises
        ------
        ValueError
            If unsupported edge types are encountered.
        """
        tasks = []
        ids = []
        for edge in edges:
            if isinstance(edge, MetadataEdge):
                tasks.append(
                    self.asearch(
                        embedding=query_embedding,
                        k=k,
                        filter=self._metadata_filter(base_filter=filter, edge=edge),
                        **kwargs,
                    )
                )
            elif isinstance(edge, IdEdge):
                ids.append(edge.id)
            else:
                raise ValueError(f"Unsupported edge: {edge}")

        if ids:
            tasks.append(self.aget(ids, filter))

        results: list[Content] = [
            c
            for completed_task in asyncio.as_completed(tasks)
            for c in await completed_task
        ]

        return top_k(
            results,
            embedding=query_embedding,
            k=k,
        )

    def _metadata_filter(
        self,
        base_filter: dict[str, Any] | None = None,
        edge: Edge | None = None,
    ) -> dict[str, Any]:
        """
        Return a filter for the `base_filter` and incoming edges from `edge`.

        Parameters
        ----------
        base_filter : dict[str, Any]
            Any base metadata filter that should be used for search.
            Generally corresponds to the user specified filters for the entire
            traversal. Should be combined with the filters necessary to support
            nodes with an *incoming* edge matching `edge`.
        edge : Edge, optional
            An optional edge which should be added to the filter.

        Returns
        -------
        :
            The metadata dictionary to use for the given filter.
        """
        metadata_filter = {**(base_filter or {})}
        assert isinstance(edge, MetadataEdge)
        if edge is None:
            metadata_filter
        else:
            metadata_filter[edge.incoming_field] = edge.value
        return metadata_filter



"""Define the base traversal strategy."""

from __future__ import annotations

import abc
import warnings
from collections.abc import Iterable
from typing import Any

from pydantic import BaseModel

from graph_retriever.types import Node


class Strategy(BaseModel, abc.ABC):
    """
    Interface for configuring node selection and traversal strategies.

    This base class defines how nodes are selected, traversed, and finalized during
    a graph traversal. Implementations can customize behaviors like limiting the depth
    of traversal, scoring nodes, or selecting the next set of nodes for exploration.

    Parameters
    ----------
    k : int, default 5
        Maximum number of nodes to retrieve during traversal.
    start_k : int, default 4
        Number of documents to fetch via similarity for starting the traversal.
        Added to any initial roots provided to the traversal.
    adjacent_k : int, default 10
        Number of documents to fetch for each outgoing edge.
    max_depth : int, optional
        Maximum traversal depth. If `None`, there is no limit.

    Attributes
    ----------
    k : int
        Maximum number of nodes to retrieve during traversal.
    start_k : int
        Number of documents to fetch via similarity for starting the traversal.
        Added to any initial roots provided to the traversal.
    adjacent_k : int
        Number of documents to fetch for each outgoing edge.
    max_depth : int
        Maximum traversal depth. If `None`, there is no limit.
    """

    k: int = 5
    start_k: int = 4
    adjacent_k: int = 10
    max_depth: int | None = None

    _query_embedding: list[float] = []

    @abc.abstractmethod
    def discover_nodes(self, nodes: dict[str, Node]) -> None:
        """
        Add discovered nodes to the strategy.

        This method updates the strategy's state with nodes discovered during
        the traversal process.

        Parameters
        ----------
        nodes : dict[str, Node]
            Discovered nodes keyed by their IDs.
        """
        ...

    @abc.abstractmethod
    def select_nodes(self, *, limit: int) -> Iterable[Node]:
        """
        Select discovered nodes to visit in the next iteration.

        This method determines which nodes will be traversed next. If it returns
        an empty list, traversal ends even if fewer than `k` nodes have been selected.

        Parameters
        ----------
        limit :
            Maximum number of nodes to select.

        Returns
        -------
        :
            Selected nodes for the next iteration. Traversal ends if this is empty.
        """
        ...

    def finalize_nodes(self, nodes: Iterable[Node]) -> Iterable[Node]:
        """
        Finalize the selected nodes.

        This method is called before returning the final set of nodes.

        Parameters
        ----------
        nodes : Iterable[Node]
            Nodes selected for finalization.

        Returns
        -------
        :
            Finalized nodes.
        """
        return nodes

    @staticmethod
    def build(
        base_strategy: Strategy,
        **kwargs: Any,
    ) -> Strategy:
        """
        Build a strategy for a retrieval operation.

        Combines a base strategy with any provided keyword arguments to
        create a customized traversal strategy.

        Parameters
        ----------
        base_strategy : Strategy
            The base strategy to start with.
        kwargs : dict, optional
            Additional configuration options for the strategy.

        Returns
        -------
        :
            A configured strategy instance.

        Raises
        ------
        ValueError
            If 'strategy' is set incorrectly or extra arguments are invalid.
        """
        # Check if there is a new strategy to use. Otherwise, use the base.
        strategy: Strategy
        if "strategy" in kwargs:
            if next(iter(kwargs.keys())) != "strategy":
                raise ValueError("Error: 'strategy' must be set before other args.")
            strategy = kwargs.pop("strategy")
            if not isinstance(strategy, Strategy):
                raise ValueError(
                    f"Unsupported 'strategy' type {type(strategy).__name__}."
                    " Must be a sub-class of Strategy"
                )
        elif base_strategy is not None:
            strategy = base_strategy
        else:
            raise ValueError("'strategy' must be set in `__init__` or invocation")

        # Warn if any of the kwargs don't exist in the strategy.
        # Note: We could rely on Pydantic with forbidden extra arguments to
        # handle this, however the experience isn't as nice (Validation error
        # rather than warning, no indication of which field, etc.).
        assert strategy is not None
        invalid_keys = _invalid_keys(strategy, kwargs)
        if invalid_keys is not None:
            warnings.warn(f"Unsupported key(s) {invalid_keys} set.")

        # Apply the kwargs to update the strategy.
        # This uses `model_validate` rather than `model_copy`` to re-apply validation.
        strategy = strategy.model_validate(
            {**strategy.model_dump(), **kwargs},
        )

        return strategy


def _invalid_keys(model: BaseModel, dict: dict[str, Any]) -> str | None:
    """
    Identify invalid keys in the given dictionary for a Pydantic model.

    Parameters
    ----------
    model : BaseModel
        The Pydantic model to validate against.
    dict : dict[str, Any]
        The dictionary to check.

    Returns
    -------
    :
        A comma-separated string of invalid keys, if any.
    """
    invalid_keys = set(dict.keys()) - set(model.model_fields.keys())
    if invalid_keys:
        return ", ".join([f"'{k}'" for k in invalid_keys])
    else:
        return None



"""Provide eager (breadth-first) traversal strategy."""

from collections.abc import Iterable

from typing_extensions import override

from graph_retriever.strategies.base import Strategy
from graph_retriever.types import Node


class Eager(Strategy):
    """
    Eager traversal strategy (breadth-first).

    This strategy selects all discovered nodes at each traversal step. It ensures
    breadth-first traversal by processing nodes layer by layer, which is useful for
    scenarios where all nodes at the current depth should be explored before proceeding
    to the next depth.

    Parameters
    ----------
    k : int, default 5
        Maximum number of nodes to retrieve during traversal.
    start_k : int, default 4
        Number of documents to fetch via similarity for starting the traversal.
        Added to any initial roots provided to the traversal.
    adjacent_k : int, default 10
        Number of documents to fetch for each outgoing edge.
    max_depth : int, optional
        Maximum traversal depth. If `None`, there is no limit.

    Attributes
    ----------
    k : int
        Maximum number of nodes to retrieve during traversal.
    start_k : int
        Number of documents to fetch via similarity for starting the traversal.
        Added to any initial roots provided to the traversal.
    adjacent_k : int
        Number of documents to fetch for each outgoing edge.
    max_depth : int
        Maximum traversal depth. If `None`, there is no limit.
    """

    _nodes: list[Node] = []

    @override
    def discover_nodes(self, nodes: dict[str, Node]) -> None:
        self._nodes.extend(nodes.values())

    @override
    def select_nodes(self, *, limit: int) -> Iterable[Node]:
        nodes = self._nodes[:limit]
        self._nodes = []
        return nodes



"""Defines the `Node` class used during graph traversal."""

from __future__ import annotations

import abc
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any, TypeAlias

from .content import Content


@dataclass
class Node:
    """
    Represents a node in the traversal graph.

    The [](`~graph_retriever.Node`) class contains information about a document
    during graph traversal, including its depth, embedding, edges, and metadata.

    Parameters
    ----------
    id : str
        The unique identifier of the document represented by this node.
    content : str
        The content.
    depth : int
        The depth (number of edges) through which this node was discovered. This
        depth may not reflect the true depth in the full graph if only a subset
        of edges is retrieved.
    embedding : list[float]
        The embedding vector of the document, used for similarity calculations.
    metadata : dict[str, Any]
        Metadata from the original document. This is a reference to the original
        document metadata and should not be modified directly. Any updates to
        metadata should be made to `extra_metadata`.
    extra_metadata : dict[str, Any]
        Additional metadata to override or augment the original document
        metadata during traversal.
    """

    id: str
    content: str
    depth: int
    embedding: list[float]
    metadata: dict[str, Any] = field(default_factory=dict)

    incoming_edges: set[Edge] = field(default_factory=set)
    outgoing_edges: set[Edge] = field(default_factory=set)

    extra_metadata: dict[str, Any] = field(default_factory=dict)


class Edge(abc.ABC):
    """
    An edge identifies properties necessary for finding matching nodes.

    Sub-classes should be hashable.
    """

    pass


@dataclass(frozen=True)
class MetadataEdge(Edge):
    """
    Link to nodes with specific metadata.

    A [](`~graph_retriever.MetadataEdge`) defines nodes with either:

    - `node.metadata[field] == value`
    - `node.metadata[field] CONTAINS value` (if the metadata is a collection).

    Parameters
    ----------
    incoming_field : str
        The name of the metadata field storing incoming edges.
    value : Any
        The value associated with the key for this edge
    """

    incoming_field: str
    value: Any


@dataclass(frozen=True)
class IdEdge(Edge):
    """
    Nodes with `node.id == id`.

    Parameters
    ----------
    id : str
        The ID of the node to link to.
    """

    id: str


@dataclass
class Edges:
    """
    Information about the incoming and outgoing edges.

    Parameters
    ----------
    incoming : set[Edge]
        Incoming edges that link to this node.
    outgoing : set[Edge]
        Edges that this node link to. These edges should be defined in terms of
        the *incoming* `Edge` they match. For instance, a link from "mentions"
        to "id" would link to `IdEdge(...)`.
    """

    incoming: set[Edge]
    outgoing: set[Edge]


EdgeFunction: TypeAlias = Callable[[Content], Edges]
"""A function for extracting edges from nodes.

Implementations should be deterministic.
"""



from graph_retriever.adapters.base import Adapter
from graph_retriever.adapters.in_memory import InMemory
from graph_retriever.content import Content
from graph_retriever.strategies.mmr import Mmr
from graph_retriever.testing.embeddings import angular_2d_embedding

from tests.testing.adapters import ANIMALS_DEPTH_0_EXPECTED, ANIMALS_QUERY
from tests.testing.invoker import SyncOrAsync


async def test_animals_keywords(animals: Adapter, sync_or_async: SyncOrAsync):
    """Test traversing a bi-directional field with lists."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("keywords", "keywords")],
        strategy=Mmr(start_k=2),
    )

    assert await traversal(k=4, max_depth=0) == ANIMALS_DEPTH_0_EXPECTED
    assert await traversal(k=4, max_depth=1) == ["cat", "gazelle", "hyena", "mongoose"]
    assert await traversal(k=6, max_depth=2) == [
        "bison",
        "cat",
        "fox",
        "gazelle",
        "hyena",
        "mongoose",
    ]


async def test_animals_habitat(animals: Adapter, sync_or_async: SyncOrAsync):
    """Test traversing a bi-directional field with singular values."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("habitat", "habitat")],
        strategy=Mmr(k=10, start_k=2),
    )

    assert await traversal(max_depth=0) == ANIMALS_DEPTH_0_EXPECTED
    assert await traversal(max_depth=1) == [
        "bobcat",
        "cobra",
        "deer",
        "elk",
        "fox",
        "mongoose",
    ]
    assert await traversal(max_depth=2) == [
        "bobcat",
        "cobra",
        "deer",
        "elk",
        "fox",
        "mongoose",
    ]


async def test_animals_habitat_to_keywords(
    animals: Adapter, sync_or_async: SyncOrAsync
):
    """Test traversing a from a singular field (habitat) to collection (keywords)."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("habitat", "keywords")],
        strategy=Mmr(k=10, start_k=2),
    )

    assert await traversal(max_depth=0) == ANIMALS_DEPTH_0_EXPECTED
    assert await traversal(max_depth=1) == ["bear", "bobcat", "fox", "mongoose"]
    assert await traversal(max_depth=2) == [
        "bear",
        "bobcat",
        "caribou",
        "fox",
        "mongoose",
    ]


async def test_angular(sync_or_async: SyncOrAsync):
    """
    Test end to end construction and MMR search.

    The embedding function used here ensures `texts` become
    the following vectors on a circle (numbered v0 through v3):

           ______ v2
          //      \\
         //        \\  v1
    v3  ||    .     || query
         \\        //  v0
          \\______//                 (N.B. very crude drawing)

    With start_k==2 and k==2, when query is at (1, ),
    one expects that v2 and v0 are returned (in some order)
    because v1 is "too close" to v0 (and v0 is closer than v1)).

    Both v2 and v3 are reachable via edges from v0, so once it is
    selected, those are both considered.
    """
    embedding = angular_2d_embedding
    v0 = Content.new("v0", "-0.124", embedding, metadata={"outgoing": "link"})
    v1 = Content.new("v1", "+0.127", embedding)
    v2 = Content.new("v2", "+0.25", embedding, metadata={"incoming": "link"})
    v3 = Content.new("v3", "+1.0", embedding, metadata={"incoming": "link"})

    traversal = sync_or_async.traverse_sorted_ids(
        query="0.0",
        store=InMemory(embedding, [v0, v1, v2, v3]),
        edges=[("outgoing", "incoming")],
        strategy=Mmr(k=2, start_k=2, max_depth=2),
    )

    assert await traversal() == ["v0", "v2"]
    assert await traversal() == ["v0", "v2"]

    # With max depth 0, no edges are traversed, so this doesn't reach v2 or v3.
    # So it ends up picking "v1" even though it's similar to "v0".
    assert await traversal(max_depth=0) == ["v0", "v1"]

    # With max depth 0 but higher `start_k`, we encounter v2
    assert await traversal(start_k=3, max_depth=0) == ["v0", "v2"]

    # v0 score is .46, v2 score is 0.16 so it won't be chosen.
    assert await traversal(score_threshold=0.2) == ["v0"]

    # with k=4 we should get all of the documents.
    assert await traversal(k=4) == ["v0", "v1", "v2", "v3"]



from .adapters import Adapter
from .content import Content
from .edges.metadata import EdgeSpec, Id
from .traversal import atraverse, traverse
from .types import Edge, EdgeFunction, Edges, IdEdge, MetadataEdge, Node

__all__ = [
    "Adapter",
    "Content",
    "Edge",
    "EdgeFunction",
    "Edges",
    "EdgeSpec",
    "Id",
    "IdEdge",
    "MetadataEdge",
    "Node",
    "traverse",
    "atraverse",
]



from graph_retriever.adapters.base import Adapter
from graph_retriever.strategies.scored import Scored
from graph_retriever.types import Node

from tests.testing.adapters import (
    ANIMALS_QUERY,
)
from tests.testing.invoker import SyncOrAsync


def score_animals(node: Node) -> float:
    return 20 - len(node.id)


async def test_animals_keywords(animals: Adapter, sync_or_async: SyncOrAsync):
    """Test traversing a bi-directional field with lists."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("keywords", "keywords")],
        strategy=Scored(scorer=score_animals, start_k=2),
    )

    # start_k=2 => 2 closest matches to the query
    assert await traversal(max_depth=0) == [
        "fox",
        "mongoose",
    ]
    # k=8, we start with 2 closest and choose 6 more with shortest names
    assert await traversal(k=8, max_depth=1) == [
        "cat",
        "coyote",
        "fox",
        "gazelle",
        "hyena",
        "jackal",
        "mongoose",
    ]
    # k=4, we start with 2 closest and choose 2 more with shortest names
    # (from "cat", "coyote", "gazelle", "hyena", "jackal")
    assert await traversal(k=4, max_depth=1) == [
        "cat",
        "fox",
        "hyena",
        "mongoose",
    ]



from collections.abc import Iterable
from typing import cast

from graph_retriever.content import Content
from graph_retriever.utils.math import cosine_similarity_top_k


def top_k(
    contents: Iterable[Content],
    *,
    embedding: list[float],
    k: int,
) -> list[Content]:
    """
    Select the top-k contents from the given contet.

    Parameters
    ----------
    contents : Iterable[Content]
        The content from which to select the top-K.
    embedding: list[float]
        The embedding we're looking for.
    k : int
        The number of items to select.

    Returns
    -------
    list[Content]
        Top-K by similarity. All results will have their `score` set.
    """
    # TODO: Consider handling specially cases of already-sorted batches (merge).
    # TODO: Consider passing threshold here to limit results.

    # Use dicts to de-duplicate by ID. This ensures we choose the top K distinct
    # content (rather than K copies of the same content).
    scored = {c.id: c for c in contents if c.score is not None}
    unscored = {c.id: c for c in contents if c.score is None if c.id not in scored}

    if unscored:
        top_unscored = _similarity_sort_top_k(
            list(unscored.values()), embedding=embedding, k=k
        )
        scored.update(top_unscored)

    sorted = list(scored.values())
    sorted.sort(key=_score, reverse=True)

    return sorted[:k]


def _score(content: Content) -> float:
    return cast(float, content.score)


def _similarity_sort_top_k(
    contents: list[Content], *, embedding: list[float], k: int
) -> dict[str, Content]:
    # Flatten the content and use a dict to deduplicate.
    # We need to do this *before* selecting the top_k to ensure we don't
    # get duplicates (and fail to produce `k`).
    top_k, scores = cosine_similarity_top_k(
        [embedding], [c.embedding for c in contents], top_k=k
    )

    results = {}
    for (_x, y), score in zip(top_k, scores):
        c = contents[y]
        c.score = score
        results[c.id] = c
    return results
